/* 

Allan Spale
CS 488
Assignment 1

This class will read the landscape and data files for the 2-D earthquake
visualization.  It will also answer data requests about extreme values
of the landscape.

*/


#include "DataManagement.h"

// buffer for reading from file
const int DataManagement::KI_BUFFER_SIZE = 100;

//----------------- PUBLIC FUNCTIONS -------------------------------//

DataManagement::DataManagement() {
	lQuakeDataSize = 10;
	lLandscapeDataSize = 10;
	iMaximumLatitude = 0;
	iMaximumLongitutde = 0;
	iMinimumLatitude = 99999;
	iMinimumLongitude = 99999;

	iLandscapeStructSize = sizeof( LandscapeLocation );
	iQuakeStructSize = sizeof( QuakeData );

	llpLandscape = new LandscapeLocation[ lLandscapeDataSize ];
	qdpQuakeData = new QuakeData[ lQuakeDataSize ];
}

		
DataManagement::~DataManagement() {
	delete [] llpLandscape;
	delete [] qdpQuakeData;
}


// returns a pointer to the earthquake data
QuakeData *
DataManagement::qdpGetEarthquakeData() {
	QuakeData *qdpFilteredData, *qdpTempFilteredData;
	long lIndex, lDataSize, lMaximumDataSizeArray;

	lMaximumDataSizeArray = 10;
	qdpFilteredData = new QuakeData[ lMaximumDataSizeArray ];
	lDataSize = 0;

	for ( lIndex = 0; lIndex < lQuakeDataSize; lIndex++ ) {
		
		memcpy( &qdpFilteredData[ lDataSize ], &qdpQuakeData[ lIndex ],
			    iQuakeStructSize );
			
		lDataSize++;

		// if there is no more room in the array,
		// double the size and copy the old space into the 
		// new space
		if ( lDataSize == lMaximumDataSizeArray ) {

			qdpTempFilteredData = 
				new QuakeData[ lQuakeDataSize * 2 ]; 

			memcpy( qdpTempFilteredData, qdpFilteredData,
					iQuakeStructSize * lMaximumDataSizeArray );
			
			delete [] qdpFilteredData;
			lMaximumDataSizeArray *= 2;
			qdpFilteredData = qdpTempFilteredData;
		}		
	}

	// trim unused space
	qdpTempFilteredData = new QuakeData[ lDataSize ];
	lQuakeDataSize = lDataSize;

	memcpy( qdpTempFilteredData, qdpFilteredData, 
			iQuakeStructSize * lDataSize );
			
	delete [] qdpFilteredData;
	qdpFilteredData = qdpTempFilteredData;

	return qdpFilteredData;
}


// returns a pointer to the landscape data
LandscapeLocation * 
DataManagement::llpGetLandscapeData() {
	LandscapeLocation *llLandscapeDataCopy;

	llLandscapeDataCopy = new LandscapeLocation[ lLandscapeDataSize ];
	memcpy( llLandscapeDataCopy, llpLandscape, 
			lLandscapeDataSize * iLandscapeStructSize );

	return llLandscapeDataCopy;
}


// read the data files
// NOTE: The method of checking for a non-existent file failed on a
// WindowsXP PC where development was done.  Every time a check
// was made for a file that did not exist, the non-existent file
// was created though no specific request was done so.  However,
// an attempt was made to use the appropriate functions for testing
// for a non-existent file.
 
bool
DataManagement::bReadDataFiles( const char *kcpLandscapeFilename,
								const char *kcpDataFilename ) {
	char c1dBuffer[ KI_BUFFER_SIZE ], cpTemp[ 10 ], cpItem[3];
	bool bValidFiles;
	long lIndex;
	LandscapeLocation *llpTempLandscape;
	QuakeData *qdpTempQuakeData;
	


	bValidFiles = false;
	strcpy( cpTemp, "         " );
	strcpy( cpItem, "  " );

	cout << "Trying to open landscape file: " << kcpLandscapeFilename << endl;
	cout << "Trying to open earthquake file: " << kcpDataFilename << endl << endl;


	// check for the files could be opened; to prevent the 
	// "short-circuit" of one error preventing both errors
	// from being shown, both files will be tested for
	// validity even if the first tested file fails
	ifstream ifLandscapeFile( kcpLandscapeFilename, ios::in );
	if ( !ifLandscapeFile ) {
		cout << "Could not open the landscape file named \""
			 << kcpLandscapeFilename << "\"" << endl << endl;
	}
	else {
		bValidFiles = true;
	}

	ifstream ifQuakeFile( kcpDataFilename, ios::in );
	if ( !ifQuakeFile ) {
		cout << "Could not open the earthquake data file named \""
			 << kcpDataFilename << "\"" << endl << endl;
	}
	else {
		bValidFiles = true;
	}


	// read data from the files and enter into appropriate arrays
	if ( bValidFiles ) {

		//-------------- LANDSCAPE DATA INPUT -------------------//
		// get landscape data first...and ignore the first line
		ifLandscapeFile.getline( c1dBuffer, KI_BUFFER_SIZE );
		lIndex = 0;

		// assuming format: latitude, longitude
		while ( ifLandscapeFile.getline( c1dBuffer, KI_BUFFER_SIZE ) ) {
			
			llpLandscape[ lIndex ].iLatitude = 
				atoi( strtok( c1dBuffer, "," ) );

			llpLandscape[ lIndex ].iLongitude = 
				atoi( strtok( NULL, "," ) );

			// check minimum and maximum values for 
			// latitude and longitude
			if ( llpLandscape[ lIndex ].iLatitude > iMaximumLatitude ) {
				iMaximumLatitude = llpLandscape[ lIndex ].iLatitude;
			}
			else if ( llpLandscape[ lIndex ].iLatitude < iMinimumLatitude ) {
				iMinimumLatitude = llpLandscape[ lIndex ].iLatitude;
			}

			if ( llpLandscape[ lIndex ].iLongitude > iMaximumLongitutde ) {
				iMaximumLongitutde = llpLandscape[ lIndex ].iLongitude;
			}
			else if ( llpLandscape[ lIndex ].iLongitude < iMinimumLongitude ) {
				iMinimumLongitude = llpLandscape[ lIndex ].iLongitude;
			}

			lIndex++;

			// if there is no more room in the array,
			// double the size and copy the old space into the 
			// new space
			
			if ( lIndex == lLandscapeDataSize ) {

				llpTempLandscape = 
					new LandscapeLocation[ lLandscapeDataSize * 2 ]; 

				memcpy( llpTempLandscape, llpLandscape,
						iLandscapeStructSize * lLandscapeDataSize );
			
				delete [] llpLandscape;
				lLandscapeDataSize *= 2;
				llpLandscape = llpTempLandscape;
			}
		} // end while

		// trim unused space
		lLandscapeDataSize = lIndex;

		llpTempLandscape = 
			new LandscapeLocation[ lLandscapeDataSize ]; 

		memcpy( llpTempLandscape, llpLandscape,
				iLandscapeStructSize * lLandscapeDataSize );
			
		delete [] llpLandscape;
		llpLandscape = llpTempLandscape;


		//-------------- QUAKE DATA INPUT -------------------//
		// get landscape data first...and ignore the first line
		ifQuakeFile.getline( c1dBuffer, KI_BUFFER_SIZE );
		lIndex = 0;

		// assuming format: Year, Month, Day, Time(hhmmss.mm)UTC,
		//				    Latitude, Longitude, Magnitude, Depth

		while ( ifQuakeFile.getline( c1dBuffer, KI_BUFFER_SIZE ) ) {
			// if not blank line, process data
			if ( strchr( c1dBuffer, ',' ) != NULL ) {
			
				qdpQuakeData[ lIndex ].cdDate.iYear = 
					atoi( strtok( c1dBuffer, "," ) );

				qdpQuakeData[ lIndex ].cdDate.iMonth = 
					atoi( strtok( NULL, "," ) );

				qdpQuakeData[ lIndex ].cdDate.iDay = 
					atoi( strtok( NULL, "," ) );

				// time processing
				strcpy( cpTemp, strtok( NULL, "," ) );
			
				qdpQuakeData[ lIndex ].tutcTime.iHour =
					atoi( strncpy( cpItem, cpTemp, 2 ) );

				qdpQuakeData[ lIndex ].tutcTime.iMinute =
					atoi( strncpy( cpItem, cpTemp + 2, 2 ) );

				qdpQuakeData[ lIndex ].tutcTime.iSecond =
					atoi( strncpy( cpItem, cpTemp + 4, 2 ) );

				qdpQuakeData[ lIndex ].tutcTime.iMillisecond =
					atoi( strncpy( cpItem, cpTemp + 7, 2 ) );

				// process remainder of line of data but
				// ** ignore Depth for this assignment **
				
				// OK to convert from double to float...ignore warnings
				qdpQuakeData[ lIndex ].qlLocation.fLatitude = 
					atof( strtok( NULL, "," ) );

				qdpQuakeData[ lIndex ].qlLocation.fLongitude = 
					atof( strtok( NULL, "," ) );

				qdpQuakeData[ lIndex ].fMagnitude = 
					atof( strtok( NULL, "," ) );

				// do not keep data that has no magnitude
				if ( qdpQuakeData[ lIndex ].fMagnitude > 0.0 ) {
					lIndex++;
				}


				// if there is no more room in the array,
				// double the size and copy the old space into the 
				// new space
				
				if ( lIndex == lQuakeDataSize ) {

					qdpTempQuakeData = 
						new QuakeData[ lQuakeDataSize * 2 ]; 

					memcpy( qdpTempQuakeData, qdpQuakeData,
							iQuakeStructSize * lQuakeDataSize );
					delete [] qdpQuakeData;
					lQuakeDataSize *= 2;
					qdpQuakeData = qdpTempQuakeData;
				}
			}  // end if
			// else, go to next line

		} // end while

		// trim unused space
		lQuakeDataSize = lIndex;
		qdpTempQuakeData = 
			new QuakeData[ lQuakeDataSize ]; 

		memcpy( qdpTempQuakeData, qdpQuakeData,
				iQuakeStructSize * lQuakeDataSize );
			
		delete [] qdpQuakeData;
		qdpQuakeData = qdpTempQuakeData;
		
	}  // end file processing
	
	return bValidFiles;
}


// return size of quake file 
long 
DataManagement::lGetQuakeDataSize() {
	return lQuakeDataSize;
}


// return size of landscape data file
long 
DataManagement::lGetLandscapeDataSize() {
	return lLandscapeDataSize;
}


// FUNCTIONS FOR RETURNING LANDSCAPE "EXTREME" VALUES
int
DataManagement::iGetMaximumLongitude() {
	return iMaximumLongitutde;
}


int
DataManagement::iGetMinimumLongitude() {
	return iMinimumLongitude;
}


int
DataManagement::iGetMaximumLatitude() {
	return iMaximumLatitude;
}


int
DataManagement::iGetMinimumLatitude() {
	return iMinimumLatitude;
}