/* 

Allan Spale
CS 488
Assignment 1

This is the main file for the running the program and rendering the data
from the earthquake and landscape files.

*/

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <GL/glut.h>
#include "DataManagement.h"
#include "CalendarDate.h"
#include "Transformation.h"
#include "Point.h"


//--------------------------WRAPPER FUNCTIONS-------------------------------//
// Originally, the OpenGL and GLUT functions were originally contained in
// the class named DataView.  However, because of compiling errors with
// being able to convert void to void (__cdecl *), these "local" functions
// will be used to call the relevant DataView public function.  Also, it is
// necessary to declare a global variable in order to call "actual" function
// from within the class.


// global vars
enum Playback { PBK_STOP, PBK_PLAY, PBK_PAUSE };

Point ptBottomLeftCorner, ptTopRightCorner;
Transformation trTransform;
float fMinimumMagnitude;
long lEarthquakeDataSize, lLandscapeDataSize, lStartingIndex, lEndingIndex;
Playback pbkAnimation;
LandscapeLocation *llpLandscapeInfo;
QuakeData *qdpQuakeInfo;
CalendarDate cdStartDate, cdEndDate;


//-------------------------NON-OpenGL FUNCTIONS------------------------------//
void zMoveDatasetWindowForward() {
	if ( lEndingIndex >= lEarthquakeDataSize ) {
		pbkAnimation = PBK_PAUSE;
	}
	else {
		cdEndDate.zGoForwardInterval();
		cdStartDate.zGoForwardInterval();

		// get earliest occurrence of the "starting date"
		while ( ( qdpQuakeInfo[ lStartingIndex ].cdDate.bIsLessThan( cdStartDate ) == true )
			&&  ( lStartingIndex < lEndingIndex ) ) {
			lStartingIndex++;
		}

		// get next occurrence of the "starting date"
		while ( ( qdpQuakeInfo[ lEndingIndex ].cdDate.bIsLessThan( cdEndDate ) == true )
			&&  ( lEndingIndex < lEarthquakeDataSize ) ) {
			lEndingIndex++; 
		}

		// get last duplicate value of the "ending date"
		while ( ( qdpQuakeInfo[ lEndingIndex ].cdDate.bIsEqualTo( cdEndDate ) == true )
			&&  ( lEndingIndex < lEarthquakeDataSize ) ) {
			lEndingIndex++; 
		}
	}

}


// step backward through the earthquake data; cannot animate backward
void zMoveDatasetWindowBackward() {
	if ( lStartingIndex > 0 ) {
		
		if ( lEndingIndex >= lEarthquakeDataSize ) {
			lEndingIndex = lEarthquakeDataSize - 1;
		}

		cdEndDate.zGoBackwardInterval();
		cdStartDate.zGoBackwardInterval();

		// get last, earliest occurrence of the "ending date"
		while ( ( qdpQuakeInfo[ lEndingIndex ].cdDate.bIsGreaterThan( cdEndDate ) == true )
			&&  ( lEndingIndex > 0 ) ) {
			lEndingIndex--; 
		}

		// get the most recent occurrence of "starting date"
		while ( ( qdpQuakeInfo[ lStartingIndex ].cdDate.bIsGreaterThan( cdStartDate ) == true )
			&&  ( lStartingIndex > 0 ) ) {
			lStartingIndex--;
		}
		// get the earliest occurrence of "starting date"
		while ( ( qdpQuakeInfo[ lStartingIndex ].cdDate.bIsEqualTo( cdStartDate ) == true )
			&&  ( lStartingIndex > 0 ) ) {
			lStartingIndex--;
		}
	}
}


// find the first occurrence of date (looking from the start of the file)
// in the earthquake data file and return the position of the
// occurrence in the file 
long cdGetLocationOfMonthInEarthquakeData( CalendarDate cdDate ) {
	long lIndex;

	lIndex = lStartingIndex;

	while ( ( cdDate.bIsGreaterThan( qdpQuakeInfo[ lIndex ].cdDate ) )  &&
			( lIndex < lEarthquakeDataSize ) ) {
		lIndex++;
	}

	return lIndex;
}
		

// process ASCII-based keypresses
void zProcessASCIIKeyboard( unsigned char ucKeypress, 
						    int iX, int iY ) {
	bool bUpdateMagnitudeFilter;
	float fPreCalculation;

	bUpdateMagnitudeFilter = true;

	switch( ucKeypress ) {


		// escape key (ASCII=27) exit program key
		case 27:
			cout << "Exiting..." << endl;
			exit(0);

		// magnitude filter keys...subtract 48 from character value to 
		// obtain integer equivalent
		case '0':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
			fMinimumMagnitude = ucKeypress - 48;
			break;


		// zoom keys...'A'/'a': zoom in,
		//			   'Z','z': zoom out
		case 'a':
		case 'A':
			trTransform.fZoomValue += trTransform.fZoomIncrement;
			break;

		case 'z':
		case 'Z':
			fPreCalculation = trTransform.fZoomValue - trTransform.fZoomIncrement;

			if ( fPreCalculation > 0.001 ) {
				trTransform.fZoomValue = fPreCalculation;
			}
			break;


		// time-travel keys...'G'/'g': time goes forward
		//                    'H'/'h': time goes backward
		case 'g':
		case 'G':
			if ( pbkAnimation == PBK_STOP ) {
				pbkAnimation = PBK_PAUSE;
				cdStartDate.zSetCalendarDate( qdpQuakeInfo[ lStartingIndex ].cdDate );
				cdEndDate.zSetCalendarDate( cdStartDate );
				cdEndDate.zGoForward30Days();
				lEndingIndex = cdGetLocationOfMonthInEarthquakeData( cdEndDate );
			}
			else if ( pbkAnimation == PBK_PLAY ) {
				pbkAnimation = PBK_PAUSE;
			}
			zMoveDatasetWindowForward();
			break;

		case 'h':
		case 'H':
			if ( pbkAnimation == PBK_STOP ) {
				pbkAnimation = PBK_PAUSE;
				cdStartDate.zSetCalendarDate( qdpQuakeInfo[ lStartingIndex ].cdDate );
				cdEndDate.zSetCalendarDate( cdStartDate );
				cdEndDate.zGoForward30Days();
				lEndingIndex = cdGetLocationOfMonthInEarthquakeData( cdEndDate );
			}
			else if ( pbkAnimation == PBK_PLAY ) {
				pbkAnimation = PBK_PAUSE;

			}
			zMoveDatasetWindowBackward();

			break;


		// data view keys...'O'/'o': show all earthquakes
		//               ...'P'/'p': playback earthquakes
		case 'o':
		case 'O':
			pbkAnimation = PBK_STOP;
			lStartingIndex = 0;
			lEndingIndex = lEarthquakeDataSize - 1;
			cdStartDate.zSetCalendarDate( qdpQuakeInfo[ lStartingIndex ].cdDate );
			cdEndDate.zSetCalendarDate( qdpQuakeInfo[ lEndingIndex ].cdDate );
			break;

		case 'p':
		case 'P':
			if ( pbkAnimation == PBK_PAUSE ) {
				// pick up from paused location
				pbkAnimation = PBK_PLAY;
			}
			else if ( pbkAnimation == PBK_STOP ) {
				// start from the beginning
				cdStartDate.zSetCalendarDate( qdpQuakeInfo[ lStartingIndex ].cdDate );
				cdEndDate.zSetCalendarDate( cdStartDate );
				cdEndDate.zGoForward30Days();
				lEndingIndex = cdGetLocationOfMonthInEarthquakeData( cdEndDate );
				pbkAnimation = PBK_PLAY;
			}
			// else ignore messge if already playing
			break;

		default:
			break;
	}
}


/*************************************************************
/* output code by Mark Kilgard */
 
void output(int x, int y, char *string)
{
   int len, i;
 
   glRasterPos2f(x, y);
   len = (int) strlen(string);
   for (i = 0; i < len; i++) {
     glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
   }
}


// process the non-ASCII based keypresses
void zProcessNonASCIIKeyboard( int iKeypress, int iX, int iY ) {
	switch( iKeypress ) {

		// pan keys...LEFT ARROW: pan left
		//			  RIGHT ARROW: pan right
		//			  UP ARROW: pan up  (note: max latitude is at top of window)
		//			  DOWN ARROW: pan down (note: min latitude is at bottom of window)
		
		case GLUT_KEY_LEFT:
			trTransform.fPanXValue += trTransform.fPanXIncrement;
			break;

		case GLUT_KEY_RIGHT:
			trTransform.fPanXValue -= trTransform.fPanXIncrement;
			break;

		case GLUT_KEY_UP:
			trTransform.fPanYValue -= trTransform.fPanYIncrement;
			break;

		case GLUT_KEY_DOWN:
			//trTransform.zIncrementPanYValue();
			trTransform.fPanYValue += trTransform.fPanYIncrement;
			break;

		default:
			break;
	}	
}


// update the visible data range for the animation
void zUpdateData( int iValue ) {
	// use for animation sequence
	if ( pbkAnimation == PBK_PLAY ) {
		zMoveDatasetWindowForward();
	}

	// keep fixed to maintain stable frame value
    glutTimerFunc(50, zUpdateData, 1);
}


// visualize the data
void zPaint() {
	long lIndex;
	float fVaryingGreenColor, fMagnitude, fXLocation, fYLocation,
			fPlotHalfSize, fXLength, fYLength, fTranslatedX, fTranslatedY,
			fPannedX, fPannedY, fPlotHalfWidthRatio;
	Point ptTextBackground[ 4 ], ptObjectPlacement[ 2 ];
	char c1dMagnitudeDisplay[ 15 ], cpDateDisplay[ 24 ];


	// calculated constants
	fPannedX = trTransform.fLandscapeOriginX + trTransform.fPanXValue;
	fPannedY = trTransform.fLandscapeOriginY + trTransform.fPanYValue;
	fPlotHalfWidthRatio = 0.025;

	ptTextBackground[ 0 ].fX = ptBottomLeftCorner.fX - 0.1;
	ptTextBackground[ 0 ].fY = ptBottomLeftCorner.fY - 0.2;

	ptTextBackground[ 1 ].fX = ptTopRightCorner.fX - 0.1;
    ptTextBackground[ 1 ].fY = ptBottomLeftCorner.fY - 0.2;

	ptTextBackground[ 2 ].fX = ptTopRightCorner.fX - 0.1;
	ptTextBackground[ 2 ].fY = ptBottomLeftCorner.fY + 0.5;

	ptTextBackground[ 3 ].fX = ptBottomLeftCorner.fX - 0.1;
	ptTextBackground[ 3 ].fY = ptBottomLeftCorner.fY + 0.5;


	ptObjectPlacement[ 0 ].fX = ptBottomLeftCorner.fX;
	ptObjectPlacement[ 0 ].fY = ptBottomLeftCorner.fY;

	ptObjectPlacement[ 1 ].fX = ptBottomLeftCorner.fX + 8.5;
	ptObjectPlacement[ 1 ].fY = ptBottomLeftCorner.fY;

	/* clear the screen to the clear colour */
    glClear(GL_COLOR_BUFFER_BIT);

	// plot earthquake data
	for ( lIndex = lStartingIndex; lIndex < lEndingIndex; lIndex++ ) {

		fMagnitude = qdpQuakeInfo[ lIndex ].fMagnitude;

		if ( fMagnitude >= fMinimumMagnitude ) {
			// determine color
			fVaryingGreenColor = 
				( 8.0 - qdpQuakeInfo[ lIndex ].fMagnitude ) / 5.0;
			glColor3f( 1.0, fVaryingGreenColor, 0.0 );

			// plot
			fYLocation = qdpQuakeInfo[ lIndex ].qlLocation.fLatitude;
			fXLocation = qdpQuakeInfo[ lIndex ].qlLocation.fLongitude;
			fPlotHalfSize = fMagnitude * fPlotHalfWidthRatio;
			
			glBegin( GL_LINES );
				// horizontal line
				fYLength = fYLocation - trTransform.fLandscapeOriginY;
				fTranslatedY = fPannedY + ( fYLength * trTransform.fZoomValue );
				
				fXLength = ( fXLocation - fPlotHalfSize ) - trTransform.fLandscapeOriginX;
				fTranslatedX = fPannedX + ( fXLength * trTransform.fZoomValue );
				glVertex2f( fTranslatedX, fTranslatedY );

				fXLength = ( fXLocation + fPlotHalfSize ) - trTransform.fLandscapeOriginX;
				fTranslatedX = fPannedX + ( fXLength * trTransform.fZoomValue );
				glVertex2f( fTranslatedX, fTranslatedY );


				// vertical line
				fXLength = fXLocation - trTransform.fLandscapeOriginX;
				fTranslatedX = fPannedX + ( fXLength * trTransform.fZoomValue );

				fYLength = ( fYLocation - fPlotHalfSize ) - trTransform.fLandscapeOriginY;
				fTranslatedY = fPannedY + ( fYLength * trTransform.fZoomValue );
				glVertex2f( fTranslatedX, fTranslatedY );
				
				fYLength = ( fYLocation + fPlotHalfSize ) - trTransform.fLandscapeOriginY;
				fTranslatedY = fPannedY + ( fYLength * trTransform.fZoomValue );
				glVertex2f( fTranslatedX, fTranslatedY );

			glEnd();
		} // end magnitude filter condition
	}
	
	// draw landscape data
	glColor3f( 1.0, 1.0, 1.0 );  //white
	
	glBegin( GL_LINE_LOOP );
		for ( lIndex = 0; lIndex < lLandscapeDataSize; lIndex++ ) {
			fYLocation = llpLandscapeInfo[ lIndex ].iLatitude;
			fXLocation = llpLandscapeInfo[ lIndex ].iLongitude;

			fXLength = llpLandscapeInfo[ lIndex ].iLongitude - trTransform.fLandscapeOriginX;
			fYLength = llpLandscapeInfo[ lIndex ].iLatitude - trTransform.fLandscapeOriginY;

			fTranslatedX = fPannedX + ( fXLength * trTransform.fZoomValue );
			fTranslatedY = fPannedY + ( fYLength * trTransform.fZoomValue );

			glVertex2f( fTranslatedX, fTranslatedY );
		}
	glEnd();

	// output text display
	
	sprintf( c1dMagnitudeDisplay, "Magnitude >= %1d", (int)fMinimumMagnitude );
	sprintf( cpDateDisplay, "%s - %s", cdStartDate.cpDisplayDate(),
		cdEndDate.cpDisplayDate() );
	
	glColor3f( 0.0, 0.0, 0.0 );
	glBegin( GL_QUADS );
		for ( lIndex = 0; lIndex < 4; lIndex++ ) {
			glVertex2f( ptTextBackground[ lIndex ].fX,
						ptTextBackground[ lIndex ].fY );
		}
	glEnd();
	glColor3f( 1.0, 0.0, 0.0 );

	
	output( ptObjectPlacement[ 0 ].fX, ptObjectPlacement[ 0 ].fY, c1dMagnitudeDisplay );
	output( ptObjectPlacement[ 1 ].fX, ptObjectPlacement[ 1 ].fY, cpDateDisplay );

	glutSwapBuffers();
}



// MAIN PROGRAM
int main( int argc, char **argv ) {
	bool bContinue;
	float fLongitudeMax, fLongitudeMin, fLatitudeMax, fLatitudeMin;
	DataManagement dmInfoStore;

	// set default values
	trTransform.fZoomIncrement = 0.01;
	trTransform.fZoomValue = 1.0;
	trTransform.fPanXIncrement = 0.1;
	trTransform.fPanYIncrement = 0.1;
	trTransform.fPanXValue = 0.0;
	trTransform.fPanYValue = 0.6;
	trTransform.fLandscapeOriginX = 0.0;
	trTransform.fLandscapeOriginY = 0.0;
	pbkAnimation = PBK_STOP;


	
	bContinue = false;
	
	if ( argc < 3 ) {
		cout << "'quakes' must have two valid filenames as parameters" << endl;
		cout << "usage: quakes landscape.txt data.txt" << endl << endl;
	}
	else {
		bContinue = 
			dmInfoStore.bReadDataFiles( argv[ 1 ], argv[ 2 ] );
	}

	if ( bContinue == true ) {

		llpLandscapeInfo = dmInfoStore.llpGetLandscapeData();
		qdpQuakeInfo = dmInfoStore.qdpGetEarthquakeData();
		lEarthquakeDataSize = dmInfoStore.lGetQuakeDataSize();
		lLandscapeDataSize = dmInfoStore.lGetLandscapeDataSize();
		
		lStartingIndex = 0;
		lEndingIndex = lEarthquakeDataSize - 1;
		
		cdStartDate.zSetCalendarDate( qdpQuakeInfo[ lStartingIndex ].cdDate );
		cdEndDate.zSetCalendarDate( qdpQuakeInfo[ lEndingIndex ].cdDate );

		// CALCULATE DAYS TO ELAPSE
		// Since 1 minute is time for animation and 50ms elapse per frame...
		// 1000 ms / 50 ms = 20 updates/sec
		// 20 updates/sec * 60 sec = 1200 updates
		//
		// roughly estimate the number of days in the earthquake data
		// by counting number of years

		CalendarDate::szSetIntervalOfDays( (int) (
			CalendarDate::slCountDaysInInterval( 
				cdStartDate, cdEndDate ) / 1200 * 2 ) );

		//----------------------SETUP OpenGL and GLUT--------------------------//
		// process any GLUT command Line options
		glutInit( &argc, argv );

		// create output window
		glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB );

		// set up the physical window size
		glutInitWindowSize( 500, 500 );

		// set the name of the window and try to create it
		glutCreateWindow("CS 488 - HW 1");

		// check for invalid colormap
		if ( glutGet( GLUT_WINDOW_COLORMAP_SIZE ) != 0 )
		{
			cout << "Error: Invalud RGBA colormap size" << endl;
			exit( 1 );
		}


		// specify function for ASCII keyboard command processing
		glutKeyboardFunc( zProcessASCIIKeyboard );

		// specify function for non-ASCII keyboard command processing
		glutSpecialFunc( zProcessNonASCIIKeyboard );

		// specify function for display
		glutDisplayFunc( zPaint );

		// specify function for execution during idle time
		glutIdleFunc( zPaint );

		// specify timer for updating the scene
		glutTimerFunc( 50, zUpdateData, 1 );


		// set up the logical graphics space
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glClearColor( 0.0, 0.0, 0.0, 1.0 );

		// order for function is left edge, right edge,
		// bottom edge, top edge
		fLongitudeMax = dmInfoStore.iGetMaximumLongitude();
		fLongitudeMin = dmInfoStore.iGetMinimumLongitude();
		fLatitudeMax = dmInfoStore.iGetMaximumLatitude();
		fLatitudeMin = dmInfoStore.iGetMinimumLatitude();

		ptBottomLeftCorner.fX = fLongitudeMin;
		ptBottomLeftCorner.fY = fLatitudeMin;
		ptTopRightCorner.fX = fLongitudeMax;
		ptTopRightCorner.fY = fLatitudeMax;

		/*trTransform.zCalculateCenterOfLandscape( fLongitudeMin, fLongitudeMax,
												 fLatitudeMin, fLatitudeMax );*/

		trTransform.fLandscapeOriginX = ( fLongitudeMax + fLongitudeMin ) / 2;
		trTransform.fLandscapeOriginY = ( fLatitudeMax + fLatitudeMin ) / 2;

		// create small border around landscape
		fLongitudeMax *= 1.005;
		fLongitudeMin *= 0.995;
		fLatitudeMax *= 1.02;
		fLatitudeMin *= 0.98;

		gluOrtho2D( fLongitudeMin, fLongitudeMax, fLatitudeMin, fLatitudeMax );

		glutMainLoop();

		return 0; 
	}
	else {
		// error messages should have already been printed by this point
		return -1;
	}
}
